/*
 * macro mlpTest.cc
 *
 * Neural Network Training interface for BaBar.  This macro can be used to 
 * train MLPs of events shapes for background discrimination.  The configuration
 * requires a function call that specifies the network configuration.
 *
 * One needs to load the MLP.  Start root and excecute the following commands.
 *
 *     gSystem->Load("libMLP.so");
 *     .x mlpTest.cc
 *
 * Need to use ROOT 4.02 or higher.
 *
 * Event Shape Variables that one can use:
 *
 * var name         variable used in NN
 *
 *  lgdr0P1n         L0, sum over neutrals particles in ROE
 *  lgdr0P1c         L0, sum over charged particles in ROE
 *  lgdr2P1n         L2, sum over neutrals particles in ROE
 *  lgdr2P1c         L2, sum over charged particles in ROE
 *  L0               L0, sum over ROE
 *  L2               L2, sum over ROE
 *  cosThetaT        |cosThetaT|
 *  catMori_idx      Moriond Tagging Category (index runs from 1-5)
 *  cosBthr          absolute value of the cosine of the B thrust axis 
 *                   with respect to the z axis
 *
 * Once the training has been completed, this macro will produce the 
 * following files:
 *
 *   nn/x.root        ROOT file containing histograms of the nn output
 *                    for signal and background; efficiencies of
 *                    signal and background and a graph of the efficiency 
 *                    of signal vs background.
 *
 *   nn/<outputfile>  an eps file summarising the output and relative strengths of
 *                    the network synapses.
 *
 *   nn/<outputfile.cxx,h>
 *                     C++ implementation required to reconstruct the neural 
 *                     network.
 *
 *   nn/*.eps          plots comparing signal vs background shapes for 
 *                     the known variables.
 *
 * The nn/ output directory is automatically created.
 *
 * More information on this set of examples can be found at the
 * following website:
 *
 *     http://www.ph.qmul.ac.uk/~bevan/yeti/
 *
 * Adrian Bevan, QMUL (C) 2007
 * e-mail: a.j.bevan@qmul.ac.uk
 */


/*
 * Function Prototypes
 */
void mlpTest();
TMultiLayerPerceptron * trainNN(TTree * tree_signal, TTree * tree_background, TString netconfig, TString outputfile);

/*
 * Global Configuration Variables
 *
 * Variable Names, Histogram Bin Numbers, Histogram Bin Ranges
 */
const Int_t NVAR = 11;

Int_t nHistBins = 100; // number of bins for nn output shape plots
Int_t nBins = 100;     // number of bins for comparison plots
Int_t nBinsCat = 5;    // number of tagging category bins

TString var_mes = "bMes";
TString var_de  = "bDeltaE";
TString var_L0n = "lgdr0P1n";
TString var_L0c = "lgdr0P1c";
TString var_L2n = "lgdr2P1n";
TString var_L2c = "lgdr2P1c";
TString var_ctt = "cosThetaT";
TString var_cbt = "cosBthr";
TString var_L0  = "L0";
TString var_L2  = "L2";
TString var_cat = "catMori";

Double_t range_mes[2] = {5.25, 5.29};
Double_t range_de[2]  = {0.2,  -0.2};
Double_t range_L0n[2] = {0.0,  10.0};
Double_t range_L0c[2] = {0.0,  10.0};
Double_t range_L2n[2] = {0.0,  10.0};
Double_t range_L2c[2] = {0.0,  10.0};
Double_t range_ctt[2] = {0.0,  1.0};
Double_t range_cbt[2] = {0.0,  1.0};
Double_t range_L0[2]  = {0.0,  10.0};
Double_t range_L2[2]  = {0.0,  10.0};
Double_t range_cat[2] = {0.0,   6.0};


/******************************************************
 * Code Implementation
 ******************************************************/
 
TMultiLayerPerceptron * trainNN(TTree * tree_signal, TTree * tree_background, TString configFileName, TString outputfile, Int_t nEpoch)
{
  TTree * TrainingSample  = new TTree("nndata", "NN Training Sample");

  Int_t nInputs = 0;
  TList variableConfig;
  TString neuralNetConfig = "";
  cout << " Reading Neural Network Configuration from: " << configFileName << endl;
  Bool_t fail = readConfigFile(configFileName, nInputs, variableConfig, neuralNetConfig);
  if(fail) return 0;

  cout << " Number of neural network inputs = " << nInputs << endl;
  cout << " Neural Network configuration    = " << neuralNetConfig << endl;
  variableConfig.Print();
  TVector variableParams(nInputs);

  Double_t bMes;
  Double_t bDeltaE;
  Double_t lgdr0P1n;
  Double_t lgdr0P1c;
  Double_t lgdr2P1n;
  Double_t lgdr2P1c;
  Double_t L0;
  Double_t L2;
  Double_t cosThetaT;
  Double_t cosBthr;

  Float_t fbMes;
  Float_t fbDeltaE;
  Float_t flgdr0P1n;
  Float_t flgdr0P1c;
  Float_t flgdr2P1n;
  Float_t flgdr2P1c;
  Float_t fL0;
  Float_t fL2;
  Float_t fcosThetaT;
  Float_t fcosBthr;

  Int_t catMori_idx;
  Int_t type;

  tree_signal->SetBranchAddress(var_mes, &bMes);
  tree_signal->SetBranchAddress(var_de,  &bDeltaE);
  tree_signal->SetBranchAddress(var_L0n, &lgdr0P1n);
  tree_signal->SetBranchAddress(var_L0c, &lgdr0P1c);
  tree_signal->SetBranchAddress(var_L2n, &lgdr2P1n);
  tree_signal->SetBranchAddress(var_L2c, &lgdr2P1c);
  tree_signal->SetBranchAddress(var_ctt, &cosThetaT);
  tree_signal->SetBranchAddress(var_cbt, &cosBthr);
  tree_signal->SetBranchAddress(var_cat + "_idx", &catMori_idx);

  tree_background->SetBranchAddress(var_mes, &bMes);
  tree_background->SetBranchAddress(var_de,  &bDeltaE);
  tree_background->SetBranchAddress(var_L0n, &lgdr0P1n);
  tree_background->SetBranchAddress(var_L0c, &lgdr0P1c);
  tree_background->SetBranchAddress(var_L2n, &lgdr2P1n);
  tree_background->SetBranchAddress(var_L2c, &lgdr2P1c);
  tree_background->SetBranchAddress(var_ctt, &cosThetaT);
  tree_background->SetBranchAddress(var_cbt, &cosBthr);
  tree_background->SetBranchAddress(var_cat + "_idx", &catMori_idx);

  TrainingSample->Branch(var_mes,      &fbMes,       var_mes + "/F");
  TrainingSample->Branch(var_de,       &fbDeltaE,    var_de + "/F");
  TrainingSample->Branch(var_L0n,      &flgdr0P1n,   var_L0n + "/F");
  TrainingSample->Branch(var_L0c,      &flgdr0P1c,   var_L0c + "/F");
  TrainingSample->Branch(var_L2n,      &flgdr2P1n,   var_L2n + "/F");
  TrainingSample->Branch(var_L2c,      &flgdr2P1c,   var_L2c + "/F");
  TrainingSample->Branch(var_ctt,      &fcosThetaT,  var_ctt + "/F");
  TrainingSample->Branch(var_cbt,      &fcosBthr,  var_cbt + "/F");
  TrainingSample->Branch(var_L0,       &fL0,         var_L0  + "/F");
  TrainingSample->Branch(var_L2,       &fL2,         var_L2  + "/F");
  TrainingSample->Branch(var_cat,      &catMori_idx, var_cat + "/I");
  TrainingSample->Branch("type",       &type,        "type/I");

  Int_t nEvents  = 2000;
  for(Int_t i=0; i<nEvents; i++)
  {
    flgdr0P1n = 0.0;
    flgdr0P1c = 0.0;
    flgdr2P1n = 0.0;
    flgdr2P1c = 0.0;
    cosThetaT = 0.0;
    bMes      = 0.0;
    bDeltaE   = 0.0;
    catMori_idx = 0;
    cosBthr = 0.0;
    tree_signal->GetEntry(i);
    type = 1;

    fcosThetaT = fabs(cosThetaT);
    fL0 = lgdr0P1n + lgdr0P1c;
    fL2 = lgdr2P1n + lgdr2P1c;
    flgdr0P1n = lgdr0P1n;
    flgdr0P1c = lgdr0P1c;
    flgdr2P1n = lgdr2P1n;
    flgdr2P1c = lgdr2P1c;
    fbMes     = bMes;
    fbDeltaE  = bDeltaE;
    fcosBthr  = cosBthr;

    TrainingSample->Fill();
  }

  for(Int_t i=0; i<nEvents; i++)
  {
    flgdr0P1n = 0.0;
    flgdr0P1c = 0.0;
    flgdr2P1n = 0.0;
    flgdr2P1c = 0.0;
    cosThetaT = 0.0;
    bMes      = 0.0;
    bDeltaE   = 0.0;
    catMori_idx = 0;
    cosBthr = 0.0;

    tree_background->GetEntry(i);
    type = 0;
    fcosThetaT = fabs(cosThetaT);
    fL0 = lgdr0P1n + lgdr0P1c;
    fL2 = lgdr2P1n + lgdr2P1c;
    flgdr0P1n = lgdr0P1n;
    flgdr0P1c = lgdr0P1c;
    flgdr2P1n = lgdr2P1n;
    flgdr2P1c = lgdr2P1c;
    fbMes     = bMes;
    fcosBthr  = cosBthr;

    TrainingSample->Fill();
  }


  cout << " >>> Printing out comparison of different variables" <<endl;
  cout << "     signal     = red" << endl;
  cout << "     background = blue" << endl;
  CompareNNInPuts(TrainingSample, var_mes);
  CompareNNInPuts(TrainingSample, var_de);
  CompareNNInPuts(TrainingSample, var_L0n);
  CompareNNInPuts(TrainingSample, var_L0c);
  CompareNNInPuts(TrainingSample, var_L2n);
  CompareNNInPuts(TrainingSample, var_L2c);
  CompareNNInPuts(TrainingSample, var_ctt);
  CompareNNInPuts(TrainingSample, var_cbt);
  CompareNNInPuts(TrainingSample, var_L0);
  CompareNNInPuts(TrainingSample, var_L2);
  CompareNNInPuts(TrainingSample, var_cat);

  TString nnMPLCode = "nn/";
  nnMPLCode += outputfile;
  nnMPLCode.ReplaceAll(".eps", "");
  cout << "About to train the nn" << endl;
  cout << "  Network Configuration = " << neuralNetConfig << endl;
  cout << "  Training sample size  = " << TrainingSample->GetEntries() << endl;
  cout << "  output MLP configuration file = " << nnMPLCode << ".*" << endl;
  TMultiLayerPerceptron * mlp = new TMultiLayerPerceptron(neuralNetConfig, TrainingSample, "Entry$%2", "Entry$/2");

  mlp->Train(nEpoch, "text,graph,update=5");

  mlp->Export(nnMPLCode);

  TCanvas* mlpa_canvas = new TCanvas("mlpa_canvas","Network analysis");
  mlpa_canvas->Divide(2,2);

  TMLPAnalyzer ana(mlp);
  // Initialisation
  ana.GatherInformations();
  // output to the console
  ana.CheckNetwork();
  mlpa_canvas->cd(1);
  // shows how each variable influences the network
  ana.DrawDInputs();

  mlpa_canvas->cd(2);
  // shows the network structure
  mlp->Draw();

  mlpa_canvas->cd(3);
  // draws the resulting network
  ana.DrawNetwork(0,"type==1","type==0");
  mlpa_canvas->cd(4);

  Double_t max = 1.5;
  Double_t min = 0.0;

  TH1F *bg  = new TH1F("bg_hist",  "NN output (background)", nHistBins, min, max);
  TH1F *sig = new TH1F("sig_hist", "NN output (signal)",     nHistBins, min, max);
  Double_t params[NVAR];
  for (Int_t i = 0; i < nEvents; i++) {
    tree_background->GetEntry(i);

    TObjString * theVar;
    TIter iter( &variableConfig );
    Int_t iInput = 0;
    while(theVar = (TObjString *)iter.Next())
    {
      if(theVar->String() == var_mes)  params[iInput] = bMes;
      if(theVar->String() == var_de)   params[iInput] = bDeltaE;
      if(theVar->String() == var_L0n)  params[iInput] = lgdr0P1n;
      if(theVar->String() == var_L0c)  params[iInput] = lgdr0P1c;
      if(theVar->String() == var_L2n)  params[iInput] = lgdr2P1n;
      if(theVar->String() == var_L2c)  params[iInput] = lgdr2P1n;
      if(theVar->String() == var_L0)   params[iInput] = lgdr0P1n + lgdr0P1c;
      if(theVar->String() == var_L2)   params[iInput] = lgdr2P1n + lgdr2P1c;
      if(theVar->String() == var_ctt)  params[iInput] = fabs(cosThetaT);
      if(theVar->String() == var_cbt)  params[iInput] = cosBthr;
      if(theVar->String() == var_cat)  params[iInput] = (Double_t)catMori_idx;

      iInput++;
    }

    Double_t nno = mlp->Evaluate(0, params);
    //if(!(i%100)) cout << "nno (bg)  = " << nno << endl;
    bg->Fill( nno );
  }
  for (Int_t i = 0; i < nEvents; i++) {
    tree_signal->GetEntry(i);

    TObjString * theVar;
    TIter iter( &variableConfig );
    Int_t iInput = 0;
    while(theVar = (TObjString *)iter.Next())
    {
      if(theVar->String() == var_mes)  params[iInput] = bMes;
      if(theVar->String() == var_de)   params[iInput] = bDeltaE;
      if(theVar->String() == var_L0n)  params[iInput] = lgdr0P1n;
      if(theVar->String() == var_L0c)  params[iInput] = lgdr0P1c;
      if(theVar->String() == var_L2n)  params[iInput] = lgdr2P1n;
      if(theVar->String() == var_L2c)  params[iInput] = lgdr2P1n;
      if(theVar->String() == var_L0)   params[iInput] = lgdr0P1n + lgdr0P1c;
      if(theVar->String() == var_L2)   params[iInput] = lgdr2P1n + lgdr2P1c;
      if(theVar->String() == var_ctt)  params[iInput] = fabs(cosThetaT);
      if(theVar->String() == var_cbt)  params[iInput] = cosBthr;
      if(theVar->String() == var_cat)  params[iInput] = (Double_t)catMori_idx;

      iInput++;
    }

    Double_t nno = mlp->Evaluate(0, params);
    //if(!(i%100)) cout << "nno (sig) = " << nno << endl;
    sig->Fill( nno );
  }
  bg->SetLineColor(kBlue);
  bg->SetFillStyle(3008);   bg->SetFillColor(kBlue);
  sig->SetLineColor(kRed);
  sig->SetFillStyle(3003); sig->SetFillColor(kRed);
  bg->SetStats(0);
  sig->SetStats(0);

  if(sig->GetMaximum() > bg->GetMaximum())
  {
    sig->Draw();
    bg->Draw("same");
  }
  else
  {
    bg->Draw();
    sig->Draw("same");
  }

  TLegend *legend = new TLegend(.75, .80, .95, .95);
  legend->AddEntry(bg, "Background");
  legend->AddEntry(sig, "Signal");
  legend->Draw();
  mlpa_canvas->cd(0);

  mlpa_canvas->Print("nn/" + outputfile);

  TH1F * sum_bg  = cumulativePlot(bg, "fromright");
  TH1F * sum_sig = cumulativePlot(sig, "fromright");
  sum_bg->SetTitle("background efficiency");
  sum_sig->SetTitle("signal efficiency");

  TCanvas c("c", "");
  sum_bg->SetLineColor(kBlue);
  sum_sig->SetLineColor(kRed);
  c.cd();
  sum_bg->Draw();
  sum_sig->Draw("same");
  c.Print("test.eps");

  sum_bg->SetName("bg_eff");
  sum_sig->SetName("sig_eff");
  sum_bg->Scale(1.0/sum_bg->GetMaximum());
  sum_sig->Scale(1.0/sum_sig->GetMaximum());

  TGraph* efficiencyPlot = getEfficiencyPlot(sum_sig, sum_bg);
  TCanvas * effcan = new TCanvas("efficiency", "canvas of signal vs background efficiency");
  efficiencyPlot->Draw("Ac");
  efficiencyPlot->GetXaxis()->SetTitle("#epsilon_{signal}");
  efficiencyPlot->GetYaxis()->SetTitle("#epsilon_{background}");
  efficiencyPlot->SetTitle("Graph of signal vs background efficiency");
  TString effoutput = outputfile;
  outputfile.ReplaceAll(".eps", "_eff.eps");
  effcan->Print("nn/" + outputfile);

  /*
   * Write out signal shape, cumulative efficiency and s_eff vs bg_eff
   * to a TFile (as well as the mlp).
   */
  outputfile.ReplaceAll(".eps", ".root");
  TFile fout("nn/" + outputfile, "RECREATE");
  fout.cd();

  bg->Write();
  sig->Write();
  sum_bg->Write();
  sum_sig->Write();
  efficiency->Write();
  efficiencyPlot->Write();

  fout.Write();
  fout.Close();

  return mlp;
}

TGraph * getEfficiencyPlot(const TH1F * hist1, const TH1F * hist2)
{
  /*
   * make a plot of eff_hist1 vs eff_hist2
   */
  Int_t nbins = hist1->GetNbinsX();
  Int_t nbinscheck = hist2->GetNbinsX();
  if(nbinscheck != nbins)
  {
    cout << "error - mismatch in number of bins used" << endl;
  }
 
  TAxis * ax = hist1->GetXaxis();
  static const Int_t NBINMAX = 5000;
  Double_t x[NBINMAX];
  Double_t y[NBINMAX];

  Double_t binlimit=0;
  for(Int_t iBin = 0; iBin < nbins; iBin++)
  {
    Double_t nsig = hist1->GetBinContent(iBin+1);
    Double_t nbg  = hist2->GetBinContent(iBin+1);
    //cout << nsig << " " << nbg << endl;
    x[iBin] = nsig;
    y[iBin] = nbg;
    if((x[iBin] == 0) && (y[iBin] == 0)) binlimit = iBin;
  }

  TGraph * graph = new TGraph(binlimit, x, y);
  graph->SetName("efficiency_graph");
  return graph;
}

void CompareNNInPuts(TTree * tree, TString variable)
{
  /*
   * Produce a plot of the MVA variable, showing signal (type = 1) and
   * background (type = 0)
   */
  system ("mkdir -p nn/fig");

  Int_t nbins = nBins;
  Double_t range * = 0;
  if(variable == var_cat)    { nbins = nBinsCat; range = range_cat; } 
  else if(variable == var_mes) range = range_mes;
  else if(variable == var_de)  range = range_de;
  else if(variable == var_L0n) range = range_L0n;
  else if(variable == var_L0c) range = range_L0c;
  else if(variable == var_L2n) range = range_L2n;
  else if(variable == var_L2c) range = range_L2c;
  else if(variable == var_L0)  range = range_L0;
  else if(variable == var_L2)  range = range_L2;
  else if(variable == var_ctt) range = range_ctt;
  else if(variable == var_cbt) range = range_cbt;

  TH1D hist_sig("signal",    "", nbins, range[0], range[1]);
  TH1D hist_bg("background", "", nbins, range[0], range[1]);

  TString DrawSig = variable;
  DrawSig        += ">>signal";
  TString DrawBg  = variable;
  DrawBg         += ">>background";
  
  tree->Draw(DrawSig, "type==1");
  tree->Draw(DrawBg, "type==0");

  hist_sig.SetLineColor(kRed);
  hist_bg.SetLineColor(kBlue);

  setAxes(&hist_sig, variable, "", "");
  setAxes(&hist_bg, variable, "", "");

  if(hist_sig.GetMaximum() > hist_bg.GetMaximum())
  {
    hist_sig.Draw();
    hist_bg.Draw("same");
  }
  else
  {
    hist_bg.Draw();
    hist_sig.Draw("same");
  }

  c1->Print("nn/fig/" + variable + ".eps");
}

Bool_t readConfigFile(TString configFileName, Int_t &nInputs, TList &variableConfig, TString &neuralNetConfig)
{
  ifstream in(configFileName.Data());
  if(!in)
  {
    cout << "unable to open config file: " << configFileName << endl;
    return kTRUE;
  }

  TString var, val;
  in >> var >> val;
  while( !in.eof() )
  {
    TObjString * VAR = new TObjString(var);
    cout << var << " " << val << endl;

    if(val == "kTRUE")
    {
      if(neuralNetConfig != "") neuralNetConfig += ",";
      variableConfig.Add(VAR);
      neuralNetConfig += var;
      nInputs++;
    }
    in >> var >> val;
  }
  neuralNetConfig += ":";
  neuralNetConfig += nInputs;
  if( (nInputs-1) > 1)
  {
    neuralNetConfig += ":";
    neuralNetConfig += nInputs-1;
  }
  neuralNetConfig += ":type";

  return kFALSE;
}


/////////////////////////////////////////////////////////////////////////////////
// given a histogram, make a new one that is the cumulative sum of this histogram
// the options that can be passed as the direction are:
//       fromleft      -  make a cumulative plot with incrasing x
//       anything else -  make a cumulative plot with decreasing x
/////////////////////////////////////////////////////////////////////////////////
TH1F * cumulativePlot(TH1 * hist, TString direction)
{
  if (!hist) return 0;

  TString name = "cumulative";
  name+= hist->GetName();
  TAxis * ax = hist->GetXaxis();
  TH1F * newHist = new TH1F(name, hist->GetTitle(), hist->GetNbinsX(), ax->GetXmin(), ax->GetXmax());

  Double_t val =0;
  if(direction == "fromleft")
    {
      cout << "cumulativePlot :: sum up the content of "<< hist->GetName()<< " from left to right"<< endl;
      for(Int_t iBin = 0; iBin <  hist->GetNbinsX(); iBin++)
	{
	  Double_t x = hist->GetBinCenter(iBin+1);
	  val+=hist->GetBinContent(iBin+1);
	  newHist->Fill(x, val);
	}
    }
  else
    {
      cout << "cumulativePlot :: sum up the content of "<< hist->GetName()<< " from right to left"<< endl;
      Int_t fillBin = 1;
      for(Int_t iBin = hist->GetNbinsX()-1; iBin >= 0; iBin--)
	{
	  Double_t x = hist->GetBinCenter(iBin+1);
	  val+=hist->GetBinContent(iBin+1);
	  newHist->Fill(x, val);
	}
    }

  return newHist;
}

void centerAxes(TH1 * hist)
{
  hist->GetXaxis()->CenterTitle();
  hist->GetYaxis()->CenterTitle();
  TH2F* h = (TH2F*)hist;
  if(h) { hist->GetZaxis()->CenterTitle(); }
}

void setTitleOffset(TH1*theHist)
{
  theHist->SetTitleOffset(1.3);
}

/***************************************************************
 * set the x,y,z titles of a histogram making sure that these  *
 * are centered as appropriate                                 *
 ***************************************************************/
void setAxes(TH1 * theHist, TString xTitle, TString yTitle, TString theTitle)
{
  centerAxes(theHist);
  setTitleOffset(theHist);

  theHist->SetXTitle(xTitle);
  theHist->SetYTitle(yTitle);
  theHist->SetTitle(theTitle);

  return;
}



void mlpTest()
{
  gROOT->SetStyle("Plain");
  /*
   * The inpput to the MLP interface is via TTrees.  Specifiy the appropriate files
   * in order to extract trees for study.
   */
  TFile signal("signal.root");
  TFile background("background.root");

  TTree * tree_signal     = (TTree*)signal.Get("data");
  TTree * tree_background = (TTree*)background.Get("data");

  if(!tree_signal)     cout << "unable to locate signal tree" << endl;
  if(!tree_background) cout << "unable to locate background tree" << endl;

  TMultiLayerPerceptron * mlp1 = trainNN(tree_signal, tree_background, "cfg/L0L2.txt", "L0L2.eps", 50);
  TMultiLayerPerceptron * mlp2 = trainNN(tree_signal, tree_background, "cfg/L0L2cosThrust.txt", "L0L2cosThrust.eps", 50);
  TMultiLayerPerceptron * mlp3 = trainNN(tree_signal, tree_background, "cfg/L0L2CosThrustCosBthr.txt", "L0L2CosThrustCosBthr.eps", 50);
}

